/*

                Create weight, trend, and exercise chart

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "pstampr.h"
#include "psrtext.h"
#include "graphics.h"
#include "bitmap.h"
#include "gifout.h"
#include "database.h"
#include "chart.h"

static RGBQUAD cmap[256];             /* Colour map */
static unsigned char *bits = NULL;    /* Bitmap */
static int cx, cy, crowbytes;         /* Bitmap dimensions */
static int ctrans;                    /* Transparent colour index or -1 */

/*  initColourMap  --  Plug our standard colours into the colour map.  */

static void initColourMap(void)
{
#define Cmap(x, r, g, b) cmap[x].rgbRed = r; cmap[x].rgbGreen = g; cmap[x].rgbBlue = b

    /*  Initialise colour map to all black.  This also guarantees the
        "rgbReserved" field is zeroed, as required.  */

    memset(cmap, 0, sizeof cmap);

    /*  Initialise standard colours.  */

    Cmap(BLACK,     0,   0,   0);
    Cmap(RED,     255,   0,   0);
    Cmap(YELLOW,  255, 255,   0);
    Cmap(GREEN,     0, 255,   0);
    Cmap(CYAN,      0, 255, 255);
    Cmap(BLUE,      0,   0, 255);
    Cmap(MAGENTA, 255,   0, 255);
    Cmap(WHITE,   255, 255, 255);
    Cmap(L_GREY,  192, 192, 192);
    Cmap(M_GREY,  128, 128, 128);
    Cmap(D_GREY,   96,  96,  96);

#undef Cmap
}

/*  initChart  --  Initialise bitmap for a new chart.  */

static int initChart(int width, int height, int bgcolour, int transcolour)
{
    assert(bits == NULL);             /* Make sure another chart isn't being built */

    cx = width;
    cy = height;
    crowbytes = ((cx + (sizeof(LONG) - 1)) / sizeof(LONG)) * sizeof(LONG);
    ctrans = transcolour;

    bits = malloc(crowbytes * cy);
    if (bits != NULL) {
        memset(bits, bgcolour, crowbytes * cy);
        initColourMap();
        psrinit(bits, width, height);
    }
    return bits != NULL;
}

/*  disposeChart  --  Discard storage allocated for current chart.  */

static void disposeChart(void)
{
    if (bits != NULL) {
        free(bits);
        bits = NULL;
    }
}

/*  exportChartGIF  --  Export the in-memory chart as a GIF file.  */

static int exportChartGIF(char *fileName)
{
    FILE *fp = fopen(fileName, "wb");
    int ok = 0;

    if (fp != NULL) {
        ok = gifout(cx, cy, crowbytes, cmap, 256, ctrans, bits, fp);
        fclose(fp);
    }
    return ok;
}

/*  editWeight  --  Edit a fixed decimal weight to a decimal value.  */

void editWeight(char *txt, weight w, int weightUnit)
{
    int whole, frac;
    
    if (weightUnit == wStone) {
        whole = (w + 50) / 100;
        sprintf(txt, "%d %2d", whole / 14, whole % 14);
    } else {
        w += 5;
        whole = w / 100;
        frac = (w / 10) % 10;
        sprintf(txt, "%d.%d", whole, frac);
    }
}


#define FntLineHeight   20            /* Font inter-line spacing */
#define FntCharHeight   10            /* Height of characters to draw */

#define leftMargin  (FntCharHeight * 5)  /* Left margin for chart */
#define rightMargin (FntCharHeight * 2) /* Right margin for chart */
#define topMargin 10                  /* Top margin for chart */
#define bottomMargin (FntCharHeight * 3) /* Bottom margin for chart */
        
#define AxisOffset  3                 /* Offset of axes from chart area */
#define TickSize    5                 /* Tick size for all axes */

/*  textWidth  --  Get width of text.  */

static int textWidth(char *s)
{
    int tboxLeft, tboxTop, tboxRight, tboxBottom;

    psr_text_box(s, FntCharHeight, 0, &tboxLeft, &tboxTop, &tboxRight, &tboxBottom);

    return tboxRight;
}

/*  drawChart  --  Draw the chart in the specified window  */

static void drawChart(struct hdMonth *m, int width, int height)
{
    int i, lday = -1;
    int pixelsPerDay, maxLabelRows;
    int lineHeight = FntLineHeight, charHgt = FntCharHeight;
    static const int factors[] = {1, 2, 5};
    int factor = 0, power = 1, vunit = 1;
    weight plotw;
    rung lrung = RUNG_ABSENT;
    weight weightMin, weightMax;      /* Weight mininum and maximum */
    trend trendMin,trendMax;          /* Trend minimum and maximum */
    struct tm today;                  /* Current date and time */
    time_t now;                       /* Current Unix second count */
    struct {
        struct {
            int x, y;
        } topLeft;
        struct {
            int x, y;
        } extent;
    } r;

    r.topLeft.x = r.topLeft.y = 0;
    r.extent.x = width;
    r.extent.y = height;
    now = time(NULL);
    today = *localtime(&now);
    today.tm_year += 1900;
    today.tm_mon++;

    /*  Find last day of month for which weight is specified.  */

    for (i = m->days - 1; i >= 0; i--) {
        if (m->weightArray[i] != 0) {
            lday = i;
            break;
        }
    }

    pixelsPerDay = (r.extent.x - (leftMargin + rightMargin)) / m->days;

    setWindowOffset(0, 0);
    setColour(BLACK);
    setFont("Roman", FntCharHeight, 0, ALIGN_LEFT | ALIGN_TOP);
        
#define PlotLine(x1, y1, x2, y2) \
                drawVector(r.topLeft.x + leftMargin + (x1),                        \
                            ((r.topLeft.y + r.extent.y) - bottomMargin) - (y1),     \
                            r.topLeft.x + leftMargin + (x2),                        \
                            ((r.topLeft.y + r.extent.y) - bottomMargin) - (y2))
                            
#define PlotRungLine(x1, y1, x2, y2) \
                drawVector(r.topLeft.x + leftMargin + (x1),                    \
                            ((r.topLeft.y + r.extent.y) - bottomMargin) - (y1),     \
                            r.topLeft.x + leftMargin + (x2),                        \
                            ((r.topLeft.y + r.extent.y) - bottomMargin) - (y2))

    /*  Fill chart area with chart background colour.  */

    setColour(D_GREY);
    fillRectangle(leftMargin + (-AxisOffset) + 1,
                  topMargin,
                  leftMargin + (pixelsPerDay * (m->days - 1)),
                  topMargin + (r.extent.y - (topMargin + bottomMargin)) + (AxisOffset + 1));
                            
    /*  Draw axes for chart  */

    setColour(BLACK);
    PlotLine(-AxisOffset, -AxisOffset, -AxisOffset, r.extent.y - (topMargin + bottomMargin));
    PlotLine(-AxisOffset, -AxisOffset, pixelsPerDay * (m->days - 1), -AxisOffset);

    /*  Label date axis  */
    
    setFont("Roman", FntCharHeight, 0, ALIGN_BASELINE | ALIGN_CENTRE);
    for (i = 0; i < (int) (m->days); i += 3) {
        char lbl[4];

        sprintf(lbl, "%d", i + 1);
        drawText(lbl,
                 r.topLeft.x + leftMargin + (pixelsPerDay * i),
                 ((r.topLeft.y + r.extent.y) - bottomMargin) + FntLineHeight);
    }

    /*  Draw ticks on date axis.  */

    for (i = 1; i < ((int) m->days); i += 1) {
        PlotLine(pixelsPerDay * i,
                 -AxisOffset,
                 pixelsPerDay * i,
                 (-AxisOffset) + TickSize);
    }
    
    /*  Plot exercise rung information */

#define RungToY(w) ((int) (((((long) (w)) - 1) * (r.extent.y - (bottomMargin + topMargin))) / 48))
    
    setColour(BLUE);
    for (i = 1; i < ((int) m->days); i++) {
        if ((m->rungArray[i - 1] != RUNG_ABSENT) && (m->rungArray[i - 1] != 0)) {
            rung rt = m->rungArray[i - 1];
            
            lrung = m->rungArray[i - 1];
            if ((m->rungArray[i] != RUNG_ABSENT) && (m->rungArray[i] != 0)) {
                rt = m->rungArray[i];
            }
            PlotRungLine(pixelsPerDay * (i - 1), RungToY(lrung),
                         pixelsPerDay * i, RungToY(rt));
        }
    }
    
    /*  Label rungs at right side of chart (draw only if any rung present) */
    
    if (lrung != RUNG_ABSENT) {
        char lrs[12];
        
        setFont("Roman", FntCharHeight, 0, ALIGN_LEFT | ALIGN_MIDDLE);
        setColour(BLACK);

        /* Plot rung axis at right side of chart. */
        PlotLine((pixelsPerDay * (m->days - 1)), -AxisOffset,
                 (pixelsPerDay * (m->days - 1)), r.extent.y - (topMargin + bottomMargin));

        for (i = 1; i <= 48; i = ((i / 6) * 6) + 6) {
            sprintf(lrs, "%d", i);
            drawText(lrs,
                         (r.topLeft.x + leftMargin + (pixelsPerDay * (m->days - 1))) + FntCharHeight,
                         ((r.topLeft.y + r.extent.y) - bottomMargin) - (RungToY(i)));
            /*  Draw ticks on rung axis.  */
            if (i > 1) {
                PlotLine((pixelsPerDay * (m->days - 1)) - TickSize, RungToY(i),
                         pixelsPerDay * (m->days - 1), RungToY(i));
            }
        }       
    }

    /*  Find weight and trend extrema for month and establish scale factors.  */
    
    weightMax = 0;
    weightMin = WEIGHT_MAX;
    trendMax = 0;
    trendMin = TREND_MAX;
    for (i = 0; i < ((int) m->days); i++) {
        weightMax = max(weightMax, m->weightArray[i]);
        if (m->weightArray[i] != 0) {
            weightMin = min(weightMin, m->weightArray[i]);
        }
        trendMax = max(trendMax, m->trendArray[i]);
        if (m->weightArray[i] != 0) {
            trendMin = min(trendMin, m->trendArray[i]);
        }
    }
    
#ifdef UNIT_CONVERSION_IMPLEMENTED
    weightMax = cvtWeight(weightMax, (weightunit) m->weightUnit, prefs.displayUnit);
    weightMin = cvtWeight(weightMin, (weightunit) m->weightUnit, prefs.displayUnit);
    trendMax = cvtTrend(trendMax, (weightunit) m->weightUnit, prefs.displayUnit);
    trendMin = cvtTrend(trendMin, (weightunit) m->weightUnit, prefs.displayUnit);
#endif

    /*  Plot weight and trend information, if any present.  */
    
    if (weightMax > 0) {
    
        if (TrendToWeight(trendMax) > weightMax) {
            weightMax = (weight) TrendToWeight(trendMax);
        }
        if (TrendToWeight(trendMin) < weightMin) {
            weightMin = (weight) TrendToWeight(trendMin);
        }
        
        maxLabelRows = (r.extent.y - (topMargin + bottomMargin)) / lineHeight;
        
        /*  Find a display scale power and factor which permits
            a suitable number of labels on the weight axis.  We do
            this using units like a vintage Tektronix oscilloscope:
            increasing powers of 1, 2, and 5.  */
            
        factor = 0;
        
        while (((weightMax - (weightMin - (weightMin % vunit))) / (factors[factor] * power)) > maxLabelRows) {
            factor++;
            if (factor > 2) {
                factor = 0;
                power *= 10;
            }
            vunit = factors[factor] * power;
        }

        /*  There's no point using a finer-grained unit than we
            plot decimal places for weight.  */

        if ((vunit < 10) && (m->weightUnit != wuStone)) {
            vunit = 10;
        }
        if ((vunit < 100) && (m->weightUnit == wuStone)) {
            vunit = 100;
        }

        /*  Round weight unit minimum to even unit multiple.  */
        
        weightMin -= weightMin % vunit;

        /*  Offset by one unit at top and bottom to avoid collision
            with axes.  */

        weightMin -= vunit;
        weightMax += vunit;
        
#ifdef UNIT_CONVERSION_IMPLEMENTED
#define WeightToY(w) ((int) (((((long) (cvtWeight(w, (weightunit) m->weightUnit, prefs.displayUnit))) - weightMin) * (r.extent.y - (bottomMargin + topMargin))) / (weightMax - weightMin)))
#else
#define WeightToY(w) ((int) (((((long) (w)) - weightMin) * (r.extent.y - (bottomMargin + topMargin))) / (weightMax - weightMin)))
#endif
#define DisplayWeightToY(w) ((int) (((((long) (w)) - weightMin) * (r.extent.y - (bottomMargin + topMargin))) / (weightMax - weightMin)))
#define TrendToY(t) (WeightToY(TrendToWeight(t)))

        /*  Plot the trend as a continuous line.  */

        setColour(RED);
        for (i = 0; i < lday; i++) {
            if ((m->weightArray[i] != 0) ||
                (m->year < today.tm_year) ||
                ((m->year == today.tm_year) && (m->month < (today.tm_mon - 1)) ||
                ((m->year == today.tm_year) && (m->month == (today.tm_mon - 1) &&
                (i <= (today.tm_mday - 1)))))) {
                UInt oy, ny;
                
                oy = TrendToY(m->trendArray[i]);
                ny = TrendToY(m->trendArray[i + 1]);
                
                PlotLine(pixelsPerDay * i, oy, pixelsPerDay * (i + 1), ny);
            }
        }
        
        /*  Plot the weights as floats and sinkers.  */
        
        for (i = 0; i < ((int) m->days); i++) {
            if (m->weightArray[i] != 0) {
                int j, ty, wy, offset;
                
                ty = TrendToY(m->trendArray[i]);
                wy = WeightToY(m->weightArray[i]);
                offset = ((int) wy) - ((int) ty);

#define SINKER  4                     /* Size of floats and sinkers */
                
                if (offset < -SINKER || offset > SINKER) {
                    int dy = offset < 0 ? -1 : 1;

                    setColour(GREEN);
                    PlotLine(pixelsPerDay * i, ty + dy,
                             pixelsPerDay * i, wy + ((offset > 0) ? -SINKER : SINKER));
                }

                /*  Fill float/sinker with white or yellow, if it's
                    flagged.  */

                setColour(m->flagsArray[i] ? YELLOW : L_GREY);
                for (j = -SINKER; j <= SINKER; j++) {
                    int dx = abs(j) - SINKER;

                    PlotLine((pixelsPerDay * i) - dx, wy + j,
                             (pixelsPerDay * i) + dx, wy + j);
                }
                
                setColour(BLUE);
                PlotLine(pixelsPerDay * i - SINKER, wy, pixelsPerDay * i, wy - SINKER);
                PlotLine(pixelsPerDay * i, wy - SINKER, pixelsPerDay * i + SINKER, wy);
                PlotLine(pixelsPerDay * i + SINKER, wy, pixelsPerDay * i, wy + SINKER);
                PlotLine(pixelsPerDay * i, wy + SINKER, pixelsPerDay * i - SINKER, wy);
            }
        }
        
        /*  Label weight axis.  */

        setFont("Roman", FntCharHeight, 0, ALIGN_RIGHT | ALIGN_MIDDLE);
        setColour(BLACK);
        for (plotw = weightMin; plotw <= weightMax; plotw += vunit) {
            char txt[12];
            
            editWeight(txt, plotw, m->weightUnit);
            drawText(txt,
                r.topLeft.x + (leftMargin - 12),
                ((r.topLeft.y + r.extent.y) - bottomMargin) - (DisplayWeightToY(plotw)));
            if (plotw > weightMin) {
                PlotLine(-AxisOffset, WeightToY(plotw),
                         (-AxisOffset) + TickSize, WeightToY(plotw));
            }
        }
    } else {
        /*  No weight data in this monthly log!  */
        
        setFont("Roman", FntCharHeight, 0, ALIGN_CENTRE | ALIGN_MIDDLE);
        setColour(BLACK);
        drawText("No weight data in this log",
                 r.topLeft.x + (r.extent.x) / 2,
                 r.topLeft.y + (r.extent.y / 2));
    }
}

/*  chartMonth  --  Prepare chart for a month.  Returns the file
                    name of the generated chart file (as a pointer
                    to a static string--multiple callers and threaded
                    programs beware) or NULL in case of error.  */

char *chartMonth(struct hdMonth *m, int width, int height,
                 int bgcolour, int transcolour)
{
    char *fn = NULL;

    if (initChart(width, height, bgcolour, transcolour)) {
        static char fname[24];

        drawChart(m, width, height);
        sprintf(fname, "%04d-%02d.gif", m->year, m->month + 1);
        if (exportChartGIF(fname)) {
            fn = fname;
        }
        disposeChart();
    }

    return fn;
}
