/*

    Import data from a Comma Separated Value (CSV) file
    
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "csv.h"
#include "database.h"

/*  parseDayOld  --  Parse daily log information from an old-format
                     (Excel) CSV record.  */

Boolean parseDayOld(char *csvLine,                     /* Input            */
                    int *year, int *month, int *day,   /* Outputs: date    */
                    weight *wgt,                       /*          weight  */
                    rung *rng,                         /*          rung    */
                    Boolean *flag,                     /*          flag    */
                    char *comment)                     /*          comment */
{
    char s[132];
    char *d1, *d2;
    char delim;
    int n1, n2, n3;
    
    *year = -1;
    *month = -1;
    *day = -1;
    *wgt = 0;
    *rng = RUNG_ABSENT;
    *flag = false;
    comment[0] = 0;
    
    CSVscanInit(csvLine);
    
    /*  Parse date.  Non-data lines are discarded at this phase  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    
    /*  We accept dates in any of the following forms:
    
            U.S.        m/d/y   m/d/yyyy
            European    d.m.yy  d.m.yyyy
            ISO 8601    yyyy-mm-dd
            ISO Idiots  yy-mm-dd
            
        Two digit dates are converted using the convention that
        70-99 map to 1970-1999 and 0 to 69 to 2000-2069.
            
    */
    
    d1 = strchr(s, '.');
    if (d1 == NULL) {
        d1 = strchr(s, '/');
    }
    if (d1 == NULL) {
        d1 = strchr(s, '-');
    }
    if (d1 == NULL) {
        return false;           /*  No date delimiter found  */
    }
    d2 = strchr(d1 + 1, *d1);
    if (d2 == NULL) {
        return false;           /*  Second data delimiter missing  */
    }
    delim = *d1;
    *d1 = *d2 = 0;
    n1 = atoi(s);
    n2 = atoi(d1 + 1);
    n3 = atoi(d2 + 1);
    switch (delim) {
        case '-':               /* ISO  */
            *year = n1;
            *month = n2;
            *day = n3;
            break;
            
        case '.':               /* European  */
            *day = n1;
            *month = n2;
            *year = n3;
            break;
            
        case '/':               /* U.S. */
            *month = n1;
            *day = n2;
            *year = n3;
            break;
    }
    if (*year < 100) {
        *year += (*year < 70) ? 2000 : 1900;
    }
    
    /*  Do some rudimentary validation of the date.  */
    
    if ((*day < 1) || (*day > DaysInMonth(*month, *year)) ||
        (*month < 1) || (*month > 12)) {
        return false;
    }
    
    /*  Field 2 is just the weekday, which we ignore.
        If it's absent, however, the record is malformed.  */
        
    if (!CSVscanField(s)) {
        return false;
    }

    /*  Field 3: weight  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if (strlen(s) > 0) {
        if (isdigit(s[0])) {
            d1 = strchr(s, '.');
            n2 = 0;
            if (d1 != NULL) {
                *d1++ = 0;
                /*  Truncate or pad as required to obtain two decimal places  */
                if (strlen(d1) > 2) {
                    d1[2] = 0;
                }
                while (strlen(d1) < 2) {
                    strcat(d1, "0");
                }
                n2 = atoi(d1);
            }
            n1 = atoi(s);
            *wgt = (n1 * 100) + n2;
        } else {
        
            /*  The convention in Excel logs was that comments
                appeared in the Weight field, to explain missing
                weights (for example, when out of town).  Thus we
                treat any entry which doesn't begin with a digit
                as a comment.  In new style logs the weight and
                comment fields are independent and both can
                appear for a given day.  */
                
            strcpy(comment, s);
        }
    }
    
    /*  Field 4: trend  This is ignored because we recompute locally  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    
    /*  Field 5: variance  This is ignored because we recompute locally  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    
    /*  Field 6: filler  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    
    /*  Field 7: rung  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if ((strlen(s) > 0) && isdigit(s[0])) {
        *rng = atoi(s);
        /*  Validate specified rung is in range  */
        if (*rng < 1 || *rng > RUNG_MAX) {
            return false;
        }
    } else {
        /*  Mark rung absent  */
        *rng = RUNG_ABSENT;
    }

    /*  Field 8: flag  */
    
    if (!CSVscanField(s)) {
        return true;
    }
    if (strlen(s) > 0) {
        *flag = (Boolean) (s[0] != '0');
    }
    return true;
}

/*  parseDay  --  Parse daily log information from an
                  Eat Watch CSV log entry.  */

Boolean parseDay(char *csvLine,                     /* Input            */
                 int *year, int *month, int *day,   /* Outputs: date    */
                 weight *wgt,                       /*          weight  */
                 rung *rng,                         /*          rung    */
                 Boolean *flag,                     /*          flag    */
                 char *comment)                     /*          comment */
{
    char s[132];
    char *d1, *d2;
    char delim;
    int n1, n2, n3;
    
    *year = -1;
    *month = -1;
    *day = -1;
    *wgt = 0;
    *rng = RUNG_ABSENT;
    *flag = false;
    comment[0] = 0;
    
    CSVscanInit(csvLine);
    
    /*  Parse date.  Non-data lines are discarded at this phase  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    
    /*  Eat watch log items always use ISO 8601 yyyy-mm-dd dates.  */
    
    d1 = strchr(s, '-');
    if (d1 == NULL) {
        return false;           /*  No date delimiter found  */
    }
    d2 = strchr(d1 + 1, *d1);
    if (d2 == NULL) {
        return false;           /*  Second data delimiter missing  */
    }
    delim = *d1;
    *d1 = *d2 = 0;
    n1 = atoi(s);
    n2 = atoi(d1 + 1);
    n3 = atoi(d2 + 1);
    *year = n1;
    *month = n2;
    *day = n3;
    
    /*  Do some rudimentary validation of the date.  */
    
    if ((*day < 1) || (*day > DaysInMonth(*month, *year)) ||
        (*month < 1) || (*month > 12)) {
        return false;
    }
    
    /*  Field 2: weight  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if (strlen(s) > 0) {
        if (isdigit(s[0])) {
            d1 = strchr(s, '.');
            n2 = 0;
            if (d1 != NULL) {
                *d1++ = 0;
                /*  Truncate or pad as required to obtain two decimal places  */
                if (strlen(d1) > 2) {
                    d1[2] = 0;
                }
                while (strlen(d1) < 2) {
                    strcat(d1, "0");
                }
                n2 = atoi(d1);
            }
            n1 = atoi(s);
            *wgt = (n1 * 100) + n2;
        } else {
            /* Old format CSV logs will contain the day of the
               week in the second field, and hence never begin
               with a digit.  Here's where we bail out if this
               record is old format. */
            return false;
        }
    }
    
    /*  Field 3: rung  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if ((strlen(s) > 0) && isdigit(s[0])) {
        *rng = atoi(s);
        /*  Validate specified rung is in range  */
        if (*rng < 1 || *rng > RUNG_MAX) {
            return false;
        }
    } else {
        /*  Mark rung absent  */
        *rng = RUNG_ABSENT;
    }

    /*  Field 4: flag  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if (strlen(s) > 0) {
        *flag = (Boolean) (s[0] != '0');
    }

    /*  Field 5: comment  */

    if (CSVscanField(s)) {
        strcpy(comment, s);
    }

    return true;
}

/*  parseHeader  --  Parse CSV header record.  */

Boolean parseHeader(char *csvLine,                  /* Input            */
                 trend *startTrend,                 /* Outputs: trendCarryover */
                 unsigned char *weightUnit,         /*          weightUnit */
                 unsigned long *lastModTime)        /*          lastModificationTime */
{
    char s[132];
    char *d1;
    int n1, n2;
    
    *startTrend = 0;
    *weightUnit = wKilogram;
    *lastModTime = 0;
    
    CSVscanInit(csvLine);
    
    /*  Field 1:  Label  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if (strcmp(s, "StartTrend") != 0) {
        return false;
    }
    
    /*  Field 2: trend carry-over  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if (strlen(s) > 0) {
        if (isdigit(s[0])) {
            d1 = strchr(s, '.');
            n2 = 0;
            if (d1 != NULL) {
                *d1++ = 0;
                /*  Truncate or pad as required to obtain four decimal places  */
                if (strlen(d1) > 4) {
                    d1[2] = 0;
                }
                while (strlen(d1) < 4) {
                    strcat(d1, "0");
                }
                n2 = atol(d1);
            }
            n1 = atoi(s);
            *startTrend = (n1 * 10000L) + n2;
        }
    }
    
    /*  Field 3: weight unit  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    if ((strlen(s) > 0) && isdigit(s[0])) {
        *weightUnit = atoi(s);
    }

    /*  Field 4: last modification time  */
    
    if (!CSVscanField(s)) {
        return false;
    }
    *lastModTime = atol(s) + timeOffset;

    return true;
}
