/*

                             H D R E A D
                             ===========

                Read Hacker's Diet Eat Watch database.

                            by John Walker
                       http://www.fourmilab.ch/
                 This program is in the public domain

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#ifdef WIN32
#include "getopt.h"
#else
#include <unistd.h>
#endif

#include "astro.h"
#include "pdb.h"
#include "palmread.h"
#include "palmrec.h"
#include "database.h"
#include "memdb.h"
#include "htmlout.h"
#include "calendar.h"
#include "cachedb.h"
#include "chart.h"
#include "version.h"
#include "xdsub.h"

int joulesp = 0;                      /* Use kilojoules as energy unit ? */

static FILE *fi;                      /* Input file */

#define CACHEFILE   "eatwatch.dbc"    /* Database cache file name */

/*  quoteCSV  --  Quote CSV field, if needed.  Note that this may return
                  a pointer to a static string, which you may need  
                  to copy, and that this isn't thread safe.  */

static char *quoteCSV(char *s)
{
    if ((strchr(s, ',') != NULL) || (strchr(s, '"') != NULL)) {
        static char qs[132];
        char *qp;
        char c;

        qp = qs;
        *qp++ = '"';

        while ((c = *s++) != 0) {
            if (strchr(s, '"') != NULL) {
                *qp++ = '"';
            }
            *qp++ = c;
        }

        *qp++ = '"';
        *qp = 0;
        return qs;
    }
    return s;
}

/*  extractMonth  --  Extract a packed month from a PDB record.  */

static struct hdMonthlyLog *extractMonth(unsigned char *buf, int bufl)
{
    int i, commentLength;
    struct palmreadContext p;
    struct hdMonthlyLog l;
    struct hdMonthlyLog *md;
    long rp;
    unsigned char b;

    palmRecParse(&p, buf, bufl);

    l.year = palmRecShort(&p);
    l.month = palmRecByte(&p);
    l.weightUnit = palmRecByte(&p);
    l.trendCarryForward = palmRecLong(&p);
    l.lastModificationTime = palmRecLong(&p);
    palmRecBytes(&p, l.flagsArray, 4);
    palmRecBytes(&p, l.rungArray, 31);
    (void) palmRecByte(&p);
    for (i = 0; i < 31; i++) {
        l.weightArray[i] = palmRecShort(&p);
    }

    /*  Save record pointer, compute length of comments,
        then restore to actually transfer the comments to
        the destination buffer.  */

    rp = p.recptr;
    while ((b = palmRecByte(&p)) != 0) {
        if (b & CompressedComment) {
            p.recptr++;
        } else {
            b =  palmRecByte(&p);
            p.recptr += b;
        }
    }
    commentLength = p.recptr - rp;

    p.recptr = rp;

    md = malloc(sizeof(struct hdMonthlyLog) + commentLength);

    if (md != NULL) {
        memcpy(md, &l, sizeof(struct hdMonthlyLog));
        palmRecBytes(&p, md->commentArray, commentLength);
    }

    return md;
}

/*  dumpMonth  --  Dump month as primate-readable text.  */

static void dumpMonth(FILE *fo, struct hdMonth *m)
{
    int i;
    char srung[4], swgt[6], strend[6], svar[6], swu[24];
    long zdo = (long) (m->lastModificationTime - timeOffset);

    strcpy(swu, weightUnits[m->weightUnit]);
    swu[0] = toupper(swu[0]);
    fprintf(fo, "\n%4d-%02d [%s]  (%d days, first day %d, start trend = %.2f %s)\n",
            m->year, m->month + 1, months[m->month], m->days, m->firstDay,
            m->trendCarryForward / 1000000.0, swu);


            fprintf(fo, "    Last modified: %s",
                asctime(gmtime(&zdo)));
    fprintf(fo, " Date  Weight  Trend    Var Rung Flg  Comment\n");
    for (i = 0; i < ((int) m->days); i++) {
        srung[0] = swgt[0] = strend[0] = svar[0] = 0;

        if (m->rungArray[i] != RUNG_ABSENT) {
            sprintf(srung, "%2d", m->rungArray[i]);
        }
        if (m->weightArray[i] != 0) {
            editWeight(swgt, m->weightArray[i], m->weightUnit);
            sprintf(svar, "%5.1f", (m->weightArray[i] / 100.0) - (m->trendArray[i] / 1000000.0));
            if (strcmp(svar, " -0.0") == 0) {
                strcpy(svar, "  0.0");
            }
            if (m->trendArray[i] != 0) {
                editWeight(strend, (weight) TrendToWeight(m->trendArray[i]), m->weightUnit);
            }
        }
        fprintf(fo, "%2d %s  %5s  %5s  %5s  %2s   %c   %s\n",
            i + 1, weekdays[(m->firstDay + i) % 7],
            swgt, strend, svar, srung,
            m->flagsArray[i] ? 'X' : ' ',
            m->commentArray[i] == NULL ? "" : m->commentArray[i]);
    }
}

/*  oldCsvMonth  --  Output month as old (Excel) CSV-format records.  */

static void oldCsvMonth(FILE *fo, struct hdMonth *m)
{
    int i;

    /* Weight needs to be long because comments go there in CSV format. */
    char srung[4], swgt[132], strend[6], svar[6], sprox[6], swu[24];
    double tcf = m->trendCarryForward / 1000000.0;

    /* CSV format doesn't understand our convention of zero
       denoting no trend carryforward.  We fill in an unspecified
       trend carryforward with the first specified weight in
       the month. */

    if (tcf == 0) {
        for (i = 0; i < ((int) m->days); i++) {
            if (m->weightArray[i] != 0) {
                tcf = m->weightArray[i] / 100.0;
                break;
            }
        }
    }
    fprintf(fo, ",,,,,,,\n");
    strcpy(swu, weightUnits[m->weightUnit]);
    swu[0] = toupper(swu[0]);
    fprintf(fo, ",,%s,,%s %d,,,\n", swu, months[m->month], m->year);
    fprintf(fo, "Trend carry forward:,,,%.2f,,,,\n",  tcf);

    sprox[0] = strend[0] = 0;
    for (i = 0; i < ((int) m->days); i++) {
        srung[0] = swgt[0] = svar[0] = 0;

        if (m->rungArray[i] != RUNG_ABSENT) {
            sprintf(srung, "%2d", m->rungArray[i]);
        }
        if (m->weightArray[i] != 0) {
            sprintf(swgt, "%.1f", m->weightArray[i] / 100.0);
            strcpy(sprox, swgt);
            sprintf(svar, "%.1f", (m->weightArray[i] / 100.0) - (m->trendArray[i] / 1000000.0));
            if (strcmp(svar, "-0.0") == 0) {
                strcpy(svar, "0.0");
            }
            if (m->trendArray[i] != 0) {
                sprintf(strend, "%.1f", m->trendArray[i] / 1000000.0);
            }
        } else {
            if (i == 0) {
        sprintf(sprox, "%.1f", tcf);
        if (strcmp(sprox, "-0.0") == 0) {
            strcpy(sprox, "0.0");
        }
                strcpy(strend, sprox);
            }
            if (m->commentArray[i] != NULL) {
                strcpy(swgt, quoteCSV(m->commentArray[i]));
            } else {
                strcpy(swgt, "---");
            }
        }
        fprintf(fo, "%d/%d/%02d,%s,%s,%s,%s,%s,%s,%s\n",
            m->month + 1, i + 1, m->year % 100, weekdayslong[(m->firstDay + i) % 7],
            swgt, strend, svar, sprox,
            srung,
            m->flagsArray[i] ? "1" : "");
    }
}

/*  csvMonth  --  Output month as new format CSV-format records.  */

static void csvMonth(struct hdMonth *m, time_t cachetime)
{
    int i;
    FILE *fo;
    char fn[24];

    /* Weight needs to be long because comments go there in CSV format. */
    char srung[4], swgt[8];

    /* Generate file name from ISO date. */
    sprintf(fn, "%04d-%02d.csv", m->year, m->month + 1);
    fo = fopen(fn, "w");
    if (fo == NULL) {
        fprintf(stderr, "Cannot create output file %s\n", fn);
        exit(1);
    }
    fprintf(fo, "Date,Weight,Rung,Flag,Comment\n");
    fprintf(fo, "StartTrend,%.4f,%d,%lu,%lu\n",
                 m->trendCarryForward / 1000000.0,
                 m->weightUnit, m->lastModificationTime - timeOffset,
                 cachetime);

    for (i = 0; i < ((int) m->days); i++) {
        srung[0] = swgt[0] = 0;

        if (m->rungArray[i] != RUNG_ABSENT) {
            sprintf(srung, "%2d", m->rungArray[i]);
        }
        if (m->weightArray[i] != 0) {
            sprintf(swgt, "%.2f", m->weightArray[i] / 100.0);
        }
        fprintf(fo, "%04d-%02d-%02d,%s,%s,%d,%s\n",
            m->year, m->month + 1, i + 1,
            swgt,
            srung,
            m->flagsArray[i],
            (m->commentArray[i] == NULL) ? "" : quoteCSV(m->commentArray[i]));
    }
    fclose(fo);
}

/*  csvCacheTime  --  Obtain cache time from new format CSV file. 
                      If the CSV file does not exist, a cache time of 0
                      is returned, which will force it to be generated.  */

static time_t csvCacheTime(int year, int month)

{
    char fn[24], s[132];
    FILE *fi;
    time_t ctime = 0;                   /* Cache time */

    /* Generate file name from ISO date. */

    sprintf(fn, "%04d-%02d.csv", year, month + 1);
    fi = fopen(fn, "r");
    if (fi != NULL) {
        int i;

        for (i = 0; i < 10; i++) {      /* Assume cache time saved in first 10 lines */
            if (fgets(s, sizeof s, fi) == NULL) {
                break;
            }
            if (strncmp(s, "StartTrend,", 11) == 0) {
                char *cp = strrchr(s, ',');

                if (cp != NULL) {
                    ctime = atol(cp + 1);
                }
                break;
            }
        }
        fclose(fi);
    }
    return ctime;
}

/*  USAGE  --  Print how-to-call information.  */

static void usage(void)
{
    fprintf(stderr, "hdread [options] ifile\n");
    fprintf(stderr, "Dump Hacker's Diet Eat Watch PDB file.\n");
    fprintf(stderr, "Release %s.\n", Version);
    fprintf(stderr, "   Options:\n");
    fprintf(stderr, "       -c             CSV format output\n");
    fprintf(stderr, "       -d             Disregard cache: unconditionally update all records\n");
    fprintf(stderr, "       -h             Output HTML log\n");
    fprintf(stderr, "       -j             Use kilojoules as energy unit\n");
    fprintf(stderr, "       -o ocsvfile    Write old (Excel) format CSV file\n");
    fprintf(stderr, "       -p hrfile      Dump as human-readable text\n");
    fprintf(stderr, "       -u             Print this message\n");
    fprintf(stderr, "       -v             Verbose: dump PDB record header information\n");
    fprintf(stderr, "       -x             Include hexadecimal dump in -p option output\n");
    fprintf(stderr, "Specify \"-\" as filename for standard output.\n");
    fprintf(stderr, "by John Walker (http://www.fourmilab.ch/)\n");
    fprintf(stderr, "This program is in the public domain.\n");
}

/*  Main program.  */

int main(int argc, char *argv[])
{
    extern char *optarg;
    extern int optind;

    int i, opt, dump = 0, html = 0, newcsv = 0, oldcsv = 0,
        usecache = 1, verbose = 0, htmlchange = 0, hexdump = 0;
    int forceCacheWrite = 0;          /* Force write of newly create cache file */
    struct palmreadContext *rc;
    PDBHeader ph;
    char *inname;
    struct hdMonthlyLog *md;
    struct hdMonth *m;
    FILE *dumpfile = stdout,
         *oldcsvfile = stdout;

    while ((opt = getopt(argc, argv, "cdhjo:p:uvx")) != -1) {
        switch (opt) {
            case 'c':
                newcsv = 1;
                break;

            case 'd':
                usecache = 0;
                break;

            case 'h':
                html = 1;
                break;

            case 'j':
                joulesp = 1;
                break;

            case 'o':
                oldcsv = 1;
                if (strcmp(optarg, "-") != 0) {
                    oldcsvfile = fopen(optarg, "w");
                    if (oldcsvfile == NULL) {
                        fprintf(stderr, "Cannot create old-format CSV output file %s.\n", optarg);
                        return 2;
                    }
                }
                break;

            case 'p':
                dump = 1;
                if (strcmp(optarg, "-") != 0) {
                    dumpfile = fopen(optarg, "w");
                    if (dumpfile == NULL) {
                        fprintf(stderr, "Cannot create human-readable output file %s.\n", optarg);
                        return 2;
                    }
                }
                break;

            case 'u':
                usage();
                return 0;

            case 'v':
                verbose = 1;
                break;

            case 'x':
                hexdump = 1;
                verbose = 1;
                break;

            case '?':
                usage();
                return 2;
        }
    }

    i = 0;
    for (; optind < argc; optind++) {
        switch (i) {
            case 0:
                inname = argv[optind];
                break;
        }
        i++;
    }

    if (hexdump && (!dump)) {
        fprintf(stderr, "Error: -x option specified without -p option also set.\n");
        usage();
        return 2;
    }

    if (verbose && (!dump)) {
        fprintf(stderr, "Error: -v option specified without -p option also set.\n");
        usage();
        return 2;
    }

    /* Error if no input file name specified. */

    if (i == 0) {
        fprintf(stderr, "No input file name specified.\n");
        usage();
        return 2;
    }

    fi = fopen(inname, "rb");
    if (fi == NULL) {
        fprintf(stderr, "Cannot open input file %s.\n", inname);
        return 2;
    }

    /*  Load the cache (if any).  */

    {
        FILE *cf = fopen(CACHEFILE, "rb");

        dbCacheOpen(cf);
        if (cf != NULL) {
            fclose(cf);
        } else {
            forceCacheWrite = 1;
        }
    }

    rc = palmReadHeader(fi, &ph);

    /*  Process records.  */

    if (rc != NULL) {
        for (i = 0; i < ph.numRecords; i++) {
            unsigned char *rdata;
            long rlen;
            PDBRecordEntry re;

            rdata = palmReadRecord(rc, i, &re, &rlen);
            if (rdata != NULL) {

                if (dump && verbose) {
                    fprintf(dumpfile, "\nRecord %d:  Length %d, Offset = %ld, Attributes 0x%X",
                           i, rlen, rc->rcp[i].offset, re.attr);
                    if (re.attr & dmRecAttrDelete) {
                        fprintf(dumpfile, " Deleted");
                    }
                    if (re.attr & dmRecAttrDirty) {
                        fprintf(dumpfile, " Dirty");
                    }
                    if (re.attr & dmRecAttrBusy) {
                        fprintf(dumpfile, " Busy");
                    }
                    if (re.attr & dmRecAttrSecret) {
                        fprintf(dumpfile, " Secret");
                    }
                    fprintf(dumpfile, ", Category %d, UniqueID %d\n",
                           re.attr & dmRecAttrCategoryMask, re.uniqueID);

                    if (hexdump) {
                        xd(dumpfile, rdata, rlen, 1);
                    }
                }

                md = extractMonth(rdata, rlen);

                /*  Add this month to the cache, marking it dirty
                    if it's new or has been modified.  */

                dbCacheAdd(md->year, md->month, rdata, rlen);

                m = unpackMonthRecord(md);

                if (dump) {
                    dumpMonth(dumpfile, m);
                }

                free(md);

                mdPutMonth(m);        /* Add month to in-memory database */
                free(rdata);
            } else {
                fprintf(stderr, "\nCannot read record %d of input file.\n", i);
                return 1;
            }
        }
        palmDisposeContext(rc);

        for (i = 0; (m = mdIndexMonth(i)) != NULL; i++) {
            struct dbMemCacheHeader *cmr = findInCache(m->year, m->month);

            assert(cmr != NULL);

            /*  HTML log and chart document.  */

            if (html) {
                time_t cacheTime = 0;

                if (usecache) {
                    cacheTime = htmlCacheTime(m->year, m->month);
                }
                if (cacheTime != cmr->h.modtime) {
#ifdef CACHEDIAG
fprintf(stderr, "HTML for %04d-%02d (cache %ld) inconsistent with cache time of %ld; generated.\n", m->year, m->month + 1, cacheTime, cmr->h.modtime);
#endif
                    htmlMonth(m, cmr->h.modtime);
                    htmlchange++;
                }
#ifdef CACHEDIAG
else {
    fprintf(stderr, "HTML for %04d-%02d consistent with cache time of %ld; not generated.\n", m->year, m->month + 1, cacheTime);
}
#endif
            }

            /*  New-format CSV database.  */

            if (newcsv) {
                time_t cacheTime = 0;

                if (usecache) {
                    cacheTime = csvCacheTime(m->year, m->month);
                }
                if (cacheTime != cmr->h.modtime) {
#ifdef CACHEDIAG
fprintf(stderr, "CSV for %04d-%02d (cache %ld) inconsistent with cache time of %ld; generated.\n", m->year, m->month + 1, cacheTime, cmr->h.modtime);
#endif
                    csvMonth(m, cmr->h.modtime);
                }
#ifdef CACHEDIAG
else {
    fprintf(stderr, "CSV for %04d-%02d consistent with cache time of %ld; not generated.\n", m->year, m->month + 1, cacheTime);
}
#endif
            }

            /*  Old (Excel macro compatible) format CSV database.  */

            if (oldcsv) {
                static int ocsvhdr = 0;

                if (!ocsvhdr) {
                    fprintf(oldcsvfile, "Date,,Weight,Trend,Variance,,Rung,Flag\n");
                    ocsvhdr = 1;
                }
                oldCsvMonth(oldcsvfile, m);
            }
        }

        /*  If we're generating HTML, create the index to months in the
            database if any HTML file was modified.  */

        if (htmlchange > 0) {
            calendarHTML("calendar.html");
        }

        mdEmptyDatabase();

        /*  Update the database cache on disc.  */

        if (forceCacheWrite || dbIsCacheDirty()) {
            FILE *cf = fopen(CACHEFILE, "wb");

            if (cf != NULL) {
                dbCacheWrite(cf);
                fclose(cf);
            }
#ifdef CACHEDIAG
dbCachePrint(stderr);
#endif
            dbCacheEmpty();
        }

    } else {
        fprintf(stderr, "Cannot read input file header.\n");
        return 1;
    }

    fclose(fi);

    return 0;
}
